#ifndef __DIR_HPP__
#define __DIR_HPP__

#include <sys/types.h>
#include <dirent.h>
#include <string>
#include <system_error>

class dir_t {
public:

	dir_t(const char * path);
	dir_t(const std::string & path);
	~dir_t();

	dir_t(dir_t &&) = delete;
	dir_t(const dir_t &) = delete;
	dir_t & operator=(dir_t &&) = delete;
	dir_t & operator=(const dir_t &) = delete;

	class iterator {
		public:

			iterator(DIR * handle);
			iterator & operator++();
			const dirent * operator->() const;
			const dirent & operator*() const;
			bool operator==(const iterator &) const;
			bool operator!=(const iterator &) const;

		private:

			DIR * m_handle;
			dirent * m_entry;
			dirent m_buffer;
	};

	iterator begin() const;
	iterator end() const;

private:

	DIR * m_handle;

	static void throw_error() __attribute__ ((noreturn));
};

inline dir_t::dir_t(const char * path)
: m_handle(opendir(path))
{
	if (m_handle == nullptr) { throw_error(); }
}

inline dir_t::dir_t(const std::string & path)
: m_handle(opendir(path.c_str()))
{
	if (m_handle == nullptr) { throw_error(); }
}

inline dir_t::~dir_t() {
	closedir(m_handle);
}

inline dir_t::iterator dir_t::begin() const {
	return iterator(m_handle);
}

inline dir_t::iterator dir_t::end() const {
	return iterator(nullptr);
}

inline dir_t::iterator::iterator(DIR * handle)
: m_handle(handle)
, m_entry(nullptr)
{
	if (m_handle != nullptr) {
		if (readdir_r(m_handle, &m_buffer, &m_entry) != 0) {
			throw std::system_error(
				std::error_code(errno, std::generic_category()),
				strerror(errno)
			);
		}
	}
}

inline dir_t::iterator & dir_t::iterator::operator++() {
	if (readdir_r(m_handle, &m_buffer, &m_entry) != 0) {
		throw std::system_error(
			std::error_code(errno, std::generic_category()),
			strerror(errno)
		);
	}
	return *this;
}

inline const dirent * dir_t::iterator::operator->() const {
	return &m_buffer;
}

inline const dirent & dir_t::iterator::operator*() const {
	return m_buffer;
}

inline bool dir_t::iterator::operator==(const iterator & it) const {
	if (m_handle == nullptr) {
		if (it.m_handle == nullptr) { return true; }
		return (it.m_entry == nullptr);
	}
	if (it.m_handle == nullptr) { return (m_entry == nullptr); }
	return (m_buffer.d_ino == it.m_buffer.d_ino);
}

inline bool dir_t::iterator::operator!=(const iterator & it) const {
	return !operator==(it);
}

inline void dir_t::throw_error() {
	throw std::system_error(
		std::error_code(errno, std::generic_category()),
		strerror(errno)
	);
}


#endif
