#include <array>
#include <memory>
#include <stdexcept>
#include <memory>
#include <cmath>
#include <curl/curl.h>
#include "license.hpp"
#include "url.hpp"
#include "string_ref.hpp"
#include "evp_context.hpp"
#include "evp_bio.hpp"
#include "util.hpp"
#include "file.hpp"

/*! Helper which automatically cleans up the CURL handle obtained from curl_easy_init() on scope exit. */
namespace std {
	template<>
	class default_delete<CURL> {
	public:

		inline void operator()(CURL * curl) const {
			curl_easy_cleanup(curl);
		}
	};
}

inline static void throw_on_error(CURLcode code) {
	if (code != CURLE_OK) {
		throw std::runtime_error(curl_easy_strerror(code));
	}
}

static size_t receive(
	const char * input,
	const size_t size,
	const size_t count,
	void * ptr
) {
	std::string * response = static_cast<std::string *>(ptr);
	const size_t length = size * count;
	const size_t current_length = response->length();
	response->reserve(current_length + length);
	response->insert(current_length, input, length);
	return length;
}

static std::string transmit(
	const std::string & postdata,
	const std::string & url
) {
	throw_on_error(curl_global_init(CURL_GLOBAL_ALL));
	std::unique_ptr<CURL> curl(curl_easy_init());
	if (!curl) {
		throw std::runtime_error("Unable to initialize CURL library");
	}
	std::string response;
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_URL, url.data()));
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_NOPROGRESS, true));
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_FOLLOWLOCATION, true));
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_WRITEFUNCTION, receive));
	//TODO Automate version numbering
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_USERAGENT, "xorcom-licensing/1.0"));
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_WRITEDATA, &response));
	throw_on_error(curl_easy_setopt(curl.get(), CURLOPT_POSTFIELDS, postdata.data()));
	throw_on_error(curl_easy_perform(curl.get()));
	curl_global_cleanup();
	return response;
}

static void assign_field(
	const string_ref & name,
	const string_ref & value,
	public_key_t & public_key,
	const std::vector<license_t *> & references
) {
	if (name == "pkey") {
		public_key = public_key_t::from_hex(value.data(), value.length());
	} else if (name.matches("token[0-9]+")) {
		const size_t index = strntoul(name.data() + 5, name.length() - 5);
		if (index < references.size()) {
			references[index]->signature = signature_t::from_hex(value.data(), value.length());
		}
	} else if (name.matches("generated[0-9]+")) {
		const size_t index = strntoul(name.data() + 9, name.length() - 9);
		if (index < references.size()) {
			references[index]->generated = microtime_t::from_string(value.data(), value.length());
		}
	} else if (name.matches("expires[0-9]+")) {
		const size_t index = strntoul(name.data() + 7, name.length() - 7);
		if (index < references.size()) {
			references[index]->expires = timestamp_t::from_string(value.data(), value.length());
		}
	} else if (name.matches("name[0-9]+")) {
		const size_t index = strntoul(name.data() + 4, name.length() - 4);
		if (index < references.size()) {
			references[index]->name.assign(value.data(), value.length());
		}
	} else if (name.matches("description[0-9]+")) {
		const size_t index = strntoul(name.data() + 11, name.length() - 11);
		if (index < references.size()) {
			references[index]->description.assign(value.data(), value.length());
		}
	}
}

inline static void invalidate_references(const std::vector<license_t *> & references) {
	for (size_t i = 0; i < references.size(); i++) {
		references[i]->signature = signature_t::invalid;
	}
}

static void process_response(
	public_key_t & public_key,
	const std::vector<license_t *> & references,
	std::string & response
) {
	invalidate_references(references);
	size_t start = 0, name_length = 0;
	char * name = nullptr;
	for (size_t i = 0, length = response.length(); i < length; i++) {
		switch (response[i]) {
			case '=': {
				name = &response[start];
				name_length = url::decode(name, name, i - start);
				start = i + 1;
				break;
			}
			case '&': {
				char * value = &response[start];
				const size_t value_length = url::decode(value, value, i - start);
				start = i + 1;
				assign_field(
					string_ref(name, name_length),
					string_ref(value, value_length),
					public_key,
					references
				);
				break;
			}
		}
	}
	if (name_length > 0) {
		char * value = &response[start];
		const size_t value_length = url::decode(value, value, response.length() - start);
		assign_field(
			string_ref(name, name_length),
			string_ref(value, value_length),
			public_key,
			references
		);
	}
}

static void encode_parameter(
	std::string & postdata,
	const std::string & name,
	const std::string & value
) {
	const std::string encoded_name(url::encode(name));
	const std::string encoded_value(url::encode(value));
	const bool append_parameter = (postdata.length() > 0);
	postdata.reserve(
		postdata.length() +
		(sizeof(char) * append_parameter) + // (optional) "&" character
		encoded_name.size() +
		sizeof(char) + // "=" character
		encoded_value.size()
	);
	if (append_parameter) { postdata.push_back('&'); }
	postdata.append(encoded_name.begin(), encoded_name.end());
	postdata.push_back('=');
	postdata.append(encoded_value.begin(), encoded_value.end());
}

inline static void save_serial_number(const std::string & serial_number) {
	file_t file(file_t::temporary(XLIC_SERIAL_FILE));
	file.write(serial_number.data(), serial_number.length());
	file.close();
	file.rename(XLIC_SERIAL_FILE);
}

static std::string determine_serial_number(
	const std::string & alternative_serial_number,
	const xlic_license_type type
) {
	std::string serial_number(dongle_serial_number());
	if (serial_number.length() == 0) {
		if (type == XLIC_VIRTUAL_LICENSE) {
			if (alternative_serial_number.length() > 0) {
				save_serial_number(alternative_serial_number);
				serial_number = alternative_serial_number;
			} else {
				serial_number = file_serial_number();
			}
		}
	}
	return serial_number;
}

static std::string build_postdata(
	const std::vector<license_t *> & references,
	const std::string & alternative_serial_number,
	const xlic_license_type type
) {
	const std::string serial_number(determine_serial_number(alternative_serial_number, type));
	std::string postdata;
	if (serial_number.length() > 0) {
		encode_parameter(postdata, "dongle", serial_number);
	}
	const std::string mac(mac_address());
	if (mac.length() > 0) {
		encode_parameter(postdata, "mac0", mac);
	} else if (type == XLIC_PHYSICAL_LICENSE) {
		throw std::logic_error("Unable to obtain MAC address");
	}
	const size_t digits = std::log10(references.size()) + 1;
	std::unique_ptr<char[]> buffer(new char[5 + digits + 1]); // "token" + digits + null-terminator
	for (size_t i = 0; i < references.size(); i++) {
		sprintf(buffer.get(), "key%zu", i);
		encode_parameter(postdata, buffer.get(), references[i]->key.to_hex());
		if (references[i]->signature == signature_t::invalid) {
			// New registrations must abort when there is no serial
			if (serial_number.length() == 0) {
				throw std::logic_error("Unable to determine serial number");
			}
		} else {
			sprintf(buffer.get(), "token%zu", i);
			encode_parameter(postdata, buffer.get(), references[i]->signature.to_hex());
		}
	}
	return postdata;
}

xlic_license * license_t::export_license() const {
	size_t memory_needed = sizeof(xlic_license);
	if (name.length() > 0) {
		memory_needed += name.length() + 1;
	}
	if (description.length() > 0) {
		memory_needed += description.length() + 1;
	}
	std::unique_ptr<xlic_license> license(
		reinterpret_cast<xlic_license *>(new char[memory_needed])
	);
	char * strings = reinterpret_cast<char *>(license.get()) + sizeof(xlic_license);
	license->type = key.type();
	license->name = nullptr;
	license->description = nullptr;
	license->generated = generated.value();
	license->expires = expires.value();
	license->key = key;
	license->signature = signature;
	if (name.length() > 0) {
		strings = (license->name = strings) + name.length() + 1;
		memcpy(license->name, name.data(), name.length() + 1);
	}
	if (description.length() > 0) {
		license->description = strings;
		memcpy(license->description, description.data(), description.length() + 1);
	}
	return license.release();
}

bool license_t::remove(const std::string & url) {
	std::vector<license_t *> references;
	references.push_back(this);
	const std::string postdata(build_postdata(references, std::string(), key.type()));
	std::string response(transmit(postdata, url));
	public_key_t public_key;
	process_response(public_key, references, response);
	if (public_key == public_key_t::invalid) {
		throw std::logic_error("Invalid response received from server");
	}
	return (response.find("revoke0=success") != std::string::npos);
}

bool license_t::valid(const public_key_t & public_key) const {
	std::string data(key.to_hex());
	data += generated.to_string();
	if (expires.value() > 0) {
		if (expires < timestamp_t::now()) {
			return false;
		}
		data += expires.to_string();
	}
	data += determine_serial_number(std::string(), key.type());
	data += mac_address();
	evp_context_t context(EVP_sha256(), evp_public_key_t::from_pem(public_key.to_pem()));
	return context.verify(data.data(), data.length(), signature);
}

bool license_t::validate(
	const std::string & url,
	public_key_t & public_key,
	const std::string & alternative_serial_number
) {
	std::vector<license_t *> references;
	references.push_back(this);
	const std::string postdata(build_postdata(references, alternative_serial_number, key.type()));
	std::string response(transmit(postdata, url));
	process_response(public_key, references, response);
	if (public_key != public_key_t::invalid) {
		return (signature != signature_t::invalid && generated != microtime_t::invalid);
	}
	throw std::logic_error("Invalid response received from server");
	return false;
}

static void validate(
	const std::string & url,
	public_key_t & public_key,
	std::list<license_t> & licenses,
	const std::string & alternative_serial_number,
	const xlic_license_type type
) {
	std::vector<license_t *> references;
	references.reserve(licenses.size());
	for (std::list<license_t>::iterator it = licenses.begin(); it != licenses.end(); ++it) {
		if (it->key.type() == type) {
			references.push_back(&*it);
		}
	}
	if (references.size() > 0) {
		const std::string postdata(build_postdata(references, alternative_serial_number, type));
		std::string response(transmit(postdata, url));
		process_response(public_key, references, response);
		if (public_key == public_key_t::invalid) {
			throw std::logic_error("Invalid response received from server");
		}
	}
}

void license_t::validate(
	const std::string & url,
	public_key_t & public_key,
	std::list<license_t> & licenses,
	const std::string & alternative_serial_number,
	const xlic_license_type type
) {
	::validate(url, public_key, licenses, alternative_serial_number, type);
}
