#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <utime.h>
#include <system_error>
#include <cstring>
#include "license_file.hpp"
#include "file.hpp"
#include "globals.hpp"

#define FILE_HEADER_MAGIC			0x43494c58 // "XLIC"
#define FILE_HEADER_VERSION			0x2

struct __attribute__((__packed__)) xlic_file_header {
	uint32_t magic;
	uint32_t version;
	uint32_t license_count;
	struct xlic_public_key public_key;
};

struct __attribute__((__packed__)) xlic_license_header {
	uint64_t generated;
	uint32_t expires;
	uint32_t data_offset;
	uint16_t name_length;
	uint16_t description_length;
	struct xlic_license_key key;
	struct xlic_signature signature;
};

license_file::license_file(const std::string & filename) {
	try {
		file_t handle(filename, O_RDONLY);
		if (handle.modified() > time(nullptr) + XLIC_MAX_CLOCK_SHIFT) {
			throw std::logic_error("System clock modification detected");
		}
		if (utime(filename.data(), nullptr) != 0) {
			throw std::system_error(
				std::error_code(errno, std::generic_category()),
				strerror(errno)
			);
		}
		const size_t filesize = handle.size();
		if (filesize < sizeof(xlic_file_header)) {
			throw std::logic_error("License file is too small");
		}
		std::vector<char> buffer(filesize);
		handle.read(buffer.data(), filesize);
		const xlic_file_header * header = reinterpret_cast<const xlic_file_header *>(buffer.data());
		if (header->magic != FILE_HEADER_MAGIC) {
			throw std::logic_error("License file magic is invalid");
		}
		if (header->version != FILE_HEADER_VERSION) {
			throw std::logic_error("License file version is invalid");
		}
		size_t datasize = sizeof(xlic_file_header) + (sizeof(xlic_license_header) * header->license_count);
		if (filesize < datasize) {
			throw std::logic_error("License count is invalid or file is truncated");
		}
		const xlic_license_header * licenses = reinterpret_cast<const xlic_license_header *>(
			buffer.data() + sizeof(xlic_file_header)
		);
		for (size_t i = 0, end = header->license_count; i < end; i++) {
			const xlic_license_header & license = licenses[i];
			const size_t offset = license.data_offset + license.name_length + license.description_length;
			datasize = (datasize < offset) ? offset : datasize;
		}
		if (datasize > filesize) {
			throw std::logic_error("License data is invalid or file is truncated");
		}
		m_public_key = public_key_t(header->public_key);
		for (size_t i = 0, end = header->license_count; i < end; i++) {
			const xlic_license_header & raw = licenses[i];
			const license_t license = {
				std::string(&buffer[raw.data_offset], raw.name_length),
				std::string(&buffer[raw.data_offset + raw.name_length], raw.description_length),
				microtime_t(raw.generated),
				timestamp_t(raw.expires),
				license_key_t(raw.key),
				signature_t(raw.signature)
			};
			m_licenses.push_back(license);
		}
	} catch (std::system_error & e) {
		if (e.code().value() != ENOENT) { throw e; }
	}
}

void license_file::add(const license_t & license) {
	m_licenses.push_back(license);
}

xlic_license_list * license_file::export_list() const {
	size_t memory_needed = sizeof(xlic_license_list) + (sizeof(xlic_license) * m_licenses.size());
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		if (it->name.length() > 0) { memory_needed += it->name.length() + 1; }
		if (it->description.length() > 0) { memory_needed += it->description.length() + 1; }
	}
	std::unique_ptr<xlic_license_list> list(
		reinterpret_cast<xlic_license_list *>(new char[memory_needed])
	);
	list->count = m_licenses.size();
	list->public_key = m_public_key;
	char * strings = reinterpret_cast<char *>(list.get()) +
		sizeof(xlic_license_list) + sizeof(xlic_license) * m_licenses.size();
	xlic_license * license = list->licenses;
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		license->type = it->key.type();
		license->name = nullptr;
		license->description = nullptr;
		license->generated = it->generated.value();
		license->expires = it->expires.value();
		license->key = it->key;
		license->signature = it->signature;
		if (it->name.length() > 0) {
			strings = (license->name = strings) + it->name.length() + 1;
			memcpy(license->name, it->name.data(), it->name.length() + 1);
		}
		if (it->description.length() > 0) {
			strings = (license->description = strings) + it->description.length() + 1;
			memcpy(license->description, it->description.data(), it->description.length() + 1);
		}
		license++;
	}
	return list.release();
}

const license_t * license_file::find(const std::string & product_name) const {
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		if (it->name == product_name) {
			if (it->valid(m_public_key)) {
				return &*it;
			}
		}
	}
	return nullptr;
}

const license_t * license_file::find(const license_key_t & license_key) const throw() {
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		if (it->key == license_key) { return &*it; }
	}
	return nullptr;
}

bool license_file::remove(
	const std::string & url,
	const license_key_t & license_key
) {
	for (std::list<license_t>::iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		if (it->key == license_key) {
			if (!it->remove(url)) {
				fprintf(stderr, "Server did not recognize the license, removing anyway\n");
			}
			m_licenses.erase(it);
			return true;
		}
	}
	return false;
}

void license_file::save(const std::string & filename) const {
	mkpath(filename);
	file_t handle(file_t::temporary(filename));
	const xlic_file_header header = {
		FILE_HEADER_MAGIC,
		FILE_HEADER_VERSION,
		static_cast<uint32_t>(m_licenses.size()),
		m_public_key
	};
	handle.write(&header, sizeof(xlic_file_header));
	uint32_t offset = sizeof(xlic_file_header) + (sizeof(xlic_license_header) * m_licenses.size());
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		const xlic_license_header license = {
			it->generated.value(),
			it->expires.value(),
			offset,
			static_cast<uint16_t>(it->name.length()),
			static_cast<uint16_t>(it->description.length()),
			it->key,
			it->signature
		};
		handle.write(&license, sizeof(xlic_license_header));
		offset += it->name.length() + it->description.length();
	}
	for (std::list<license_t>::const_iterator it = m_licenses.begin(); it != m_licenses.end(); ++it) {
		handle.write(it->name.data(), it->name.length());
		handle.write(it->description.data(), it->description.length());
	}
	handle.close();
	handle.rename(filename);
}

static void erase_invalid_licenses(std::list<license_t> & licenses) {
	for (std::list<license_t>::iterator it = licenses.begin(); it != licenses.end();) {
		if (it->signature == signature_t::invalid) {
			fprintf(
				stderr,
				"License %s for %s is invalid\n",
				it->key.to_base64().data(),
				it->name.data()
			);
			it = licenses.erase(it);
		} else {
			++it;
		}
	}
}

void license_file::validate(
	const std::string & url,
	const std::string & alternative_serial_number,
	const std::string & filename
) {
	// Validate files locally and drop invalidated licenses
	for (std::list<license_t>::iterator it = m_licenses.begin(); it != m_licenses.end();) {
		if (!it->valid(m_public_key)) {
			fprintf(
				stderr,
				"License %s for %s is invalid\n",
				it->key.to_base64().data(),
				it->name.data()
			);
			it = m_licenses.erase(it);
		} else {
			++it;
		}
	}
	save(filename);
	// Validate against server
	license_t::validate(url, m_public_key, m_licenses, alternative_serial_number, XLIC_VIRTUAL_LICENSE);
	erase_invalid_licenses(m_licenses);
	save(filename);
	license_t::validate(url, m_public_key, m_licenses, std::string(), XLIC_PHYSICAL_LICENSE);
	erase_invalid_licenses(m_licenses);
	save(filename);
}
