#include "url.hpp"
#include "globals.hpp"

/*! Maco for determining whether a character is unreserved or not,
* used in calculate_output_length() and xlic_urlencode_direct().
* @param [in] character: Character to evaluate.
* @return True if the character is not reserved, false if otherwise. */
inline static bool unreserved_character(const char character) throw() {
	return ((character >= 'A' && character <= 'Z') ||
		(character >= 'a' && character <= 'z') ||
		(character >= '0' && character <= '9') ||
		(character == '-') ||
		(character == '_') ||
		(character == '.') ||
		(character == '~'));
}

/*! Determine the binary length of the URL-encoded input.
* @param [in] input: Pointer to block to be decoded.
* @param [in] input: Number of bytes to be decoded.
* @return Number of bytes required to hold the result. */
size_t url::decoded_length(
	const void * input,
	const size_t length
) throw() {
	size_t percent_signs = 0;
	const char * in = static_cast<const char *>(input);
	for (size_t i = 0; i < length; i++) {
		if (in[i] == '%') { percent_signs++; }
	}
	return length - (percent_signs * 2);
}

size_t url::decode(
	void * output,
	const void * input,
	size_t length
) throw() {
	const uint8_t * in = static_cast<const uint8_t *>(input);
	char * out = static_cast<char *>(output);
	for (size_t i = 0; i < length; i++) {
		if (in[i] == '%' && i + 2 < length) {
			uint8_t byte = hex2bin_map[in[i + 1]] << 4;
			*out++ = byte | hex2bin_map[in[i + 2]];
			i += 2;
		} else if (in[i] =='+') {
			*out++ = ' ';
		} else {
			*out++ = in[i];
		}
	}
	return out - static_cast<char *>(output);
}

/*! Determine the binary length of the URL-encoded input.
* @param [in] input: Pointer to block to be decoded.
* @param [in] input: Number of bytes to be decoded.
* @return Number of bytes required to hold the result. */
size_t url::encoded_length(
	const void * input,
	const size_t length
) throw() {
	size_t output_length = length;
	const char * in = static_cast<const char *>(input);
	for (size_t i = 0; i < length; i++) {
		if (!unreserved_character(*in++)) {
			output_length += 2; // %NN requires three bytes
		}
	}
	return output_length;
}

size_t url::encode(
	void * output,
	const void * input,
	size_t length
) throw() {
	const uint8_t * in = static_cast<const uint8_t *>(input);
	char * out = static_cast<char *>(output);
	for (size_t i = 0; i < length; i++) {
		const uint8_t character = *in++;
		if (unreserved_character(character)) {
			*out++ = character;
		} else if (character == ' ') {
			*out = '+';
		} else {
			*out++ = '%';
			*out++ = bin2hex_map[character >> 4];
			*out++ = bin2hex_map[character & 0xf];
		}
	}
	*out = 0;
	return out - static_cast<char *>(output);
}
