#include "util.hpp"
#include "globals.hpp"
#include "file.hpp"
#include "dir.hpp"
#include <auth-dongle/api.h>
#include <iostream>
#include <stdio.h>
#include <map>
#include <sys/stat.h>

static const size_t mac_file_size = 18;

uint64_t strntoull(
	const void * input,
	const size_t length
) {
	const char * in = static_cast<const char *>(input);
	uint64_t result = 0;
	for (size_t i = 0; i < length; i++) {
		const char c = *in++;
		if (c >= '0' && c <= '9') {
			result *= 10;
			result += (c - '0');
		}
	}
	return result;
}

std::string file_serial_number() {
	try {
		file_t file(XLIC_SERIAL_FILE, O_RDONLY);
		const size_t filesize = file.size();
		std::string serial_number(filesize, 0);
		file.read(&serial_number[0], filesize);
		return serial_number;
	} catch (std::system_error & e) {
		if (e.code().value() != ENOENT) {
			throw e;
		}
	}
	return std::string();
}

#define BUF_SIZE 50 // Length of auth dongle response: 28 bytes

/** Get the serial number from the hardware dongle using auth-dongle */
std::string dongle_serial_number() {
#ifdef XLIC_OVERRIDE_DONGLE_RESULT
	return std::string(XLIC_OVERRIDE_DONGLE_RESULT);
#else
	struct auth_dongle *auth_dongle;
	static const char *serial = NULL;

	/* Assume that it won't change for the runtime of the program.
	 * Getting it is expensive. So we better cache the result. */
	if (serial)
		return std::string(serial);

	auth_dongle  = auth_dongle_setup(NULL, NULL, 3, 500);	/* retries + timeout */
	if (!auth_dongle)
		return std::string("");	// No-dongle -- empty serial

	serial = auth_dongle_read_serial(auth_dongle);
	auth_dongle_cleanup(auth_dongle);
	return std::string(serial);
#endif // XLIC_OVERRIDE_DONGLE_RESULT
}

std::string mac_address() {
	dir_t net("/sys/class/net/");
	std::map<int, std::string> devices;
	char buffer[64];
	for (auto it = net.begin(), end = net.end(); it != end; ++it) {
		if (it->d_type == DT_LNK) {
			{
				std::string filename("/sys/class/net/");
				filename.append(it->d_name);
				filename.append("/type");
				file_t file(filename, O_RDONLY);
				const int len = file.try_read(buffer, sizeof(buffer));
				if (len == -1) { throw_system_error(); }
				const size_t type = strntoull(buffer, len);
				if (type != 1) { continue; }
			}
			{
				std::string filename("/sys/class/net/");
				filename.append(it->d_name);
				filename.append("/ifindex");
				file_t file(filename, O_RDONLY);
				const int len = file.try_read(buffer, sizeof(buffer));
				if (len == -1) { throw_system_error(); }
				const size_t index = strntoull(buffer, len);
				devices[index] = it->d_name;
			}
		}
	}
	for (auto it = devices.begin(), end = devices.end(); it != end; ++it) {
		std::string filename("/sys/class/net/");
		filename.append(it->second);
		filename.append("/address");
		file_t file(filename, O_RDONLY);
		const int len = file.try_read(buffer, sizeof(buffer));
		if (len == -1) { throw_system_error(); }
		return std::string(buffer, len - 1); // has trailing newline
	}
	return std::string();
}

void throw_system_error() {
	throw std::system_error(
		std::error_code(errno, std::generic_category()),
		strerror(errno)
	);
}

void mkpath(const std::string & filename) {
	size_t ofs = filename.find('/', 1);
	while (ofs != std::string::npos) {
		const std::string path = filename.substr(0, ofs);
		if (mkdir(path.c_str(), 0755) == -1) {
			if (errno != EEXIST) { throw_system_error(); }
		}
		ofs = filename.find('/', ofs + 1);
	}
}
